/*
 * Copyright 2000 by Jean-Claude Christophe <jch@oleane.net>
 *                                          <jch@bsdfr.org>
 *
 * Home page:  http://www.bsdfr.org/projects/tac_bsd/
 *
 * login.c  This file contains functions used when the user is authenticated
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ttyent.h>
#include <paths.h>
#include <syslog.h>
#include <utmp.h>
#include <pwd.h>
#include <strings.h>

#include "tac_bsd.h"
#include "conf.h"

#define LOCAL 1
#define REMOTE 2

/* in connect.c */
extern int connect_to_srv(char *);

/* in packet.c */
extern int send_authen(int, char*, char *, char *);
extern int send_auth_cont(int, char *, int);
extern int read_authen(int);
extern int read_result(int);

/* in main.c */
extern void close_quit(void);

uid_t uid_default = 272;
struct passwd *pwd; /* if user is local */
char *salt;
extern int done, try;
int login_status;   /* local or remote. used by login() in order to know how to proceed */

void
login()
{
    char *home = (char *)malloc(strlen(directory) + 1 + strlen(username)); /* +1 for the '/' */
    char tbuf[MAXPATHLEN + 2];
    char *p;

    sprintf(home, "%s/%s", directory, username);

    /* if you remove these lines, uid and gid will be 0 by default */
    if (login_status == LOCAL) {
	setegid(pwd->pw_gid);
	setgid(pwd->pw_gid);
	initgroups(username, pwd->pw_gid);
	seteuid(pwd->pw_uid);
	setuid(pwd->pw_uid);
    } else {
	if ((seteuid(uid_default) < 0) || (setuid(uid_default) < 0)) {
	    if (slog)
		syslog(_syslog, "in %s, uid failed. Login not accepted.", __FUNCTION__);
	    close_quit();
	}
    }

    setenv("TERM", term, 1);
    setenv("LOGNAME", username, 1);
    setenv("USER", username, 1);
    setenv("PATH", _PATH_DEFPATH, 1);
    
    if (login_status == LOCAL) {
	setenv("HOME", pwd->pw_dir, 1);
	setenv("SHELL", pwd->pw_shell, 1);
	setlogin(username);
	tbuf[0] = '-';
	strncpy(tbuf + 1, (p = strrchr(pwd->pw_shell, '/')) ? p + 1 : pwd->pw_shell, sizeof tbuf - 1 - 1);
	tbuf[sizeof tbuf - 1] = '\0';
	chdir(pwd->pw_dir);
	execlp(shell, tbuf, 0);
    } else {
	setenv("HOME", home, 1);
	setenv("SHELL", shell, 1);
	setlogin(username);
	execlp(shell, "", 0);
    }
}

/* from the login.c of bsd */
char *
stypeof(ttyid)
    char *ttyid;
{
    struct ttyent *t;

    return (ttyid && (t = getttynam(ttyid)) ? t->ty_type : "unknown");
}

/* 
 * returns 1 if the user is in the local password DB
 * -1 in other case.
 */
int
local_user()
{
    pwd = getpwnam(username);
    if (pwd)
	return(1);

    return(-1);
}

void
login_local()
{
    salt = pwd->pw_passwd;
    if (pwd->pw_uid == 0) {
	if (slog)
	    syslog(_syslog, "in %s, try to log in root on tty %s. Denied.", __FUNCTION__, port);
	try++;
	return;
    }

    /* check the password */
    if (strcmp(crypt(password, salt), pwd->pw_passwd)) {
	if (slog)
	    syslog(_syslog, "in %s, bad password for user %s on tty %s", __FUNCTION__, username, port);
	try++;
	return;
    }
    done = 1; /* good local authentication */
    login_status = LOCAL;

    endpwent(); /* we close open files */
}

void
login_remote()
{
    int srv_sd;   /* socket used by the server */
    int result;

    /* First step: connection and START request */
    srv_sd = connect_to_srv(server);
    
    if (srv_sd < 0) 
	exit(srv_sd);
    
    if (send_authen(srv_sd, username, password, port) < 0) {
	if (slog) 
	    fprintf(stderr, "Error in first step: check into syslog for details\n");
	close_quit();
    }
    
    /* Second step: wait the response, we're expecting a GETPASS */
    if (read_authen(srv_sd) < 0) {
	if (slog) 
	    fprintf(stderr, "Error in second step: check into syslog for details\n");
	close_quit();
    }
    
    /* Third step: send the CONTINUE packet */
    if (send_auth_cont(srv_sd, password, strlen(password)) < 0) {
	if (slog) 
	    fprintf(stderr, "Error in third step: check into syslog for details\n");
	close_quit();
    }
    
    /* Fourth step: get the fatal result: PASS or FAIL */
    result = read_result(srv_sd);
    if (result < 0) {
	if (slog) 
	    fprintf(stderr, "Error in fourth step: check into syslog for details\n");
	close_quit();
    } else if (result) {
	try++;
	printf("\nAuthentication failed.\n\n");
    } else {
	done = 1; /* done */
	login_status = REMOTE;
    }
    shutdown(srv_sd, SHUT_RDWR); close(srv_sd);
}
